% Experiments with meta interpreters
%
% Credit to Markus Triska (triska@gmx.at) who wrote a page
% about Prolog meta-interpreters:
%  http://stud4.tuwien.ac.at/~e0225855/acomip/acomip.html
%
% Features
%
% 		depth limit
%		iterative deepening
%		all solutions predicates with existential qualification
%		forall
%		negation as failure
%
%	to do
%		breadth first
% 		mode checking
% 		heuristics for goal reordering (like mode checks)
:- module(meta,[
		mmi/3
	,	mi/1
	,	mirl/2
	,	miit/1
	,	midl/1
	,	check_bindings/1
	,	wrap_existential/3
	,	on_backtracking/1
	,	find_n/4
	,	find_nmax/4
	,	cut_after/2
	]).

:- use_module(library(utils)).

:- module_transparent mmi/3, find_n/4, cut_after/2, on_backtracking/1.

% meta-meta interpreter - handles common goals for any given meta interpreter
:- index(mmi(0,1,0)).

mmi(_, true, true) :- !.
mmi(_, !   , on_backtracking(throw(cut))) :- !.
mmi(G/MIG, \+G, \+MIG) :- !.
mmi(G/MIG, once(G), once(MIG)) :- !.

mmi(G/MIG, findall(H,G,S), findall(H,MIG,S)) :- !.
mmi(G/MIG, setof(H,P,S), setof(H,F,S))       :- !, wrap_existential(G/MIG,P,F).
mmi(G/MIG, bagof(H,G,S), bagof(H,MIG,S))     :- !.
mmi(G/MIG, forall(P,Q), forall(MIP,MIQ))     :- !, 
	copy_term(G/MIG,P/MIP), 
	copy_term(G/MIG,Q/MIQ). 

% exceptions are wrapped with meta/1 so they conflict with any
% other expections generated by the meta-interpreter itself, such
% as those used to hand cut.
mmi(_, throw(X), throw(meta(X))) :- !.
mmi(G/MIG, catch(P,X,Q), catch(MIP,meta(X),MIQ)) :- !,
	copy_term(G/MIG,P/MIP), 
	copy_term(G/MIG,Q/MIQ). 

mmi(G/MIG,(T->P;Q),(MIT->MIP;MIQ)) :- !, 
	copy_term(G/MIG,T/MIT), 
	copy_term(G/MIG,P/MIP), 
	copy_term(G/MIG,Q/MIQ). 

mmi(G/MIG,(P;Q),(MIP;MIQ)) :- !, 
	copy_term(G/MIG,P/MIP), 
	copy_term(G/MIG,Q/MIQ). 

mmi(G/MIG,(P,Q),(MIP,MIQ)) :- !, 
	copy_term(G/MIG,P/MIP), 
	copy_term(G/MIG,Q/MIQ). 

mmi(_,A,A)    :- predicate_property(A,built_in), !.

on_backtracking(_).
on_backtracking(P) :- P, !, fail.


% vanilla meta-circular meta-interpreter that handles all-solns predicates and cut
% using mmi to do all the work.
mi(G) :- mmi(Z/mi(Z),G,F), !, F.
mi(G) :- catch((clause(G,Body),mi(Body)),cut,fail).

wrap_existential(F,X^P,X^Q) :- !, wrap_existential(F,P,Q).
wrap_existential(P/T,P,T).

% meta-interpreter with recursion limit
% seems not to be meta-circular for some reason
% What we really ought to do is report the recursion depth reached
% and whether or not we stopped because the limit was reached.
% this will help iterative deepening realise when to stop.

mirl(N,G) :- mmi(Z/mirl(N,Z),G,F), !, F.
mirl(N,G) :- succ(M,N),
	catch((clause(G,Body),mirl(M,Body)),cut,fail).

% iterative deepening
miit(G) :- between(0,inf,N), mirl(N,G).


/*
 * More explicit object language
 */


% This version uses first argument tagging instead of cut
% Requires Horn clauses to be written in a different way.
mit(G) :- mixlate(G,H), miti(H).
miti(true).
miti(fail)      :- fail.
miti((A,B))     :- miti(A), miti(B). 
miti(goal(G))   :- mi_clause(G,Body), miti(Body).
miti(opaque(G)) :- call(G).

% This version can defer goals if they look tricky.
mid(G) :- mixlate(G,H), midi(H).
midi(true).
midi(fail)      :- fail.
midi(goal(G))   :- mi_clause(G,Body), midi(Body).
midi(opaque(G)) :- call(G).
midi((A,B))     :- 
	defer(A) -> 
		midi(B), midi(A)
	;	midi(A), midi(B).

/*
 * List based object language
 */


% This version uses a new object language made of lists
mil(G) :- phrase(milg(G),GX), mili(GX).
mili([]).
mili([goal(G)|GX])   :- mil_clause(G,Body), mili(Body), mili(GX).
mili([opaque(G)|GX]) :- call(G), mili(GX).

% This version uses difference list clause expansion
% in order to prepend the clauses subgoals to the current list of
% goals. The interepreter is tail recursive.
% Also added clauses for all solutions predicates
midl(G) :- phrase(milg(G),GX), midli(GX).
midli([]).
midli([goal(G)|GX])   :- midl_clause(G,GXX-GX), midli(GXX).
midli([opaque(G)|GX]) :- call(G), midli(GX).
midli([all(V,H,G,S)|GX]) :- call(V,H,midli(G),S), midli(GX).
midli([forall(P,Q)|GX]) :- forall(midli(P),midli(Q)), midli(GX).
midli([not(G)|GX]) :- \+midli(G), midli(GX).

% This version builds an explicit list of or-alternatives and
% could therefore support control of backtracking 
mib(G) :- mibi([[G]-G],G).
mibi([[]-G|_],G).
mibi(Alts,G) :-
	mibs(Alts,Alts1),
	mibi(Alts1,G).

% this could be implemented in different ways to do different
% kinds of search, eg breadth first. This version doesn't
% deal with built in predicates though.
mibs([A|Ax],Ay) :-
	% A is the first alternative. It is a list of conjucts
	% together with the original goal G.
	% We take the first conjuct G0, find all of it's expansions,
	% and produce Gxx, which is a list of all the possible
	% expanded conjuct lists.
	findall(Gx-G,(A=[G0|Rest]-G,midl_clause(G0,Gx-Rest)),Gxx),
	% We then replace the original alternative A with the new
	% list of alternatives Gxx. Since it goes at the head of
	% the list, this is a depth first strategy.
	append(Gxx,Ax,Ay).


% Interpreter ops
:- op(1200,xfx,/?/).

% The term on the left determines which meta interpreter is used
(rl(N)   /?/ G) :- mirl(N,G).
(t       /?/ G) :- mit(G).


/*
 * testing
 */

%mi_clause(nat(0),true).
%mi_clause(nat(s(X)),goal(nat(X))).
nat(0).
nat(s(X)) :- nat(X).

xy(0).
xy(x(X)) :- xy(X).
xy(y(X)) :- xy(X).

% for testing cut interpretation
describe(N,small) :- N<3, !.
describe(N,medium) :- N<8, !.
describe(_,large).


% difference list clauses
midl_clause_db(nat(0),GX-GX).
midl_clause_db(nat(s(X)),[goal(nat(X))|GX]-GX).

% deferral based on instantiation of arguments
defer(opaque(succ(A,B))) :- var(A), var(B).


/*
 * Clause translators
 */

% simple goal tagging
mi_clause(G,Body) :- clause(G,B1), mixlate(B1,Body).

mixlate(true,true)     :- !.
mixlate((A,B),(AA,BB)) :- !, mixlate(A,AA), mixlate(B,BB).
mixlate(A,opaque(A))   :- predicate_property(A,built_in), !. 
mixlate(A,goal(A)).

% goal list
mil_clause(G,GX) :- clause(G,Body), phrase(milg(Body),GX).

milg(true)  --> !, [].
milg((A,B)) --> !, milg(A), milg(B).
milg(!)     --> !, [cut].
milg(\+P)   --> !, { phrase(milg(P),MP) }, [not(MP)].
milg(findall(X,P,S)) --> !, { phrase(milg(P),MP) }, [all(findall,X,MP,S)].
milg(setof(X,P,S))   --> !, { phrase(milg(P),MP) }, [all(setof,X,MP,S)].
milg(bagof(X,P,S))   --> !, { phrase(milg(P),MP) }, [all(bagof,X,MP,S)].
milg(forall(P,Q))    --> !, { phrase(milg(P),MP), phrase(milg(Q),MQ)}, [forall(MP,MQ)].
milg(A)     --> { predicate_property(A,built_in)}, !, [opaque(A)]. 
milg(A)     --> [goal(A)].

% difference list
midl_clause(G,GX-Y) :- midl_clause_db(G,GX-Y).
midl_clause(G,GX-Y) :- 
	\+midl_clause_db(G,_), % only translate if no explicit clauses
	clause(G,Body), 
	phrase(milg(Body),GX,Y). % use difference list version of phrase

% Stuff for open world meta interpreter
%		op(1000,xfy,and)
%	,	op(1100,xfy,or)
%	,	op(1200,fx,prove)
%	,	and/2, or/2		% aliases for , and ;
%	,	(prove)/1		% open world meta-interpreter
%prove(P) :- P, write('* True'), nl, !.
%prove(P) :- false(P), write('* False'), nl, !.
%prove(P) :- \+P, \+false(P), write('* Unknown'), nl.
%user:false(false(P)) :- P.
%user:false((A,B)) :- false(A); false(B).
%user:false((A;B)) :- false(A), false(B).


check_bindings([]) :- nl. % !! Do we always want this?
check_bindings([B|BT]) :-
	current_prolog_flag(toplevel_print_options,Opts),
	nl, write_bindings(Opts,[B|BT]),
	get_key([';','\r','.'],K), (K='\r';K='.'), nl. 

write_bindings(Opts,[N=V]) :- format('  ~w = ~@ ',[N,write_term(V,Opts)]).
write_bindings(Opts,[N=V,X|T]) :- 
	format('  ~w = ~@\n',[N,write_term(V,Opts)]),
	write_bindings(Opts,[X|T]).
	

% find_n and cut_after - finding the first N solutions to a goal
:- dynamic '$found'/2.

% cut_after(N:natural,Q:goal).
%
% Succeeds upto N times on backtracking
cut_after(N,Query) :-
	gensym('$find',ID), % unique ID for this call

	% the idea is that we use a dynamic predicate to
	% carry information through backtracking. We simply
	% need to keep track of the number of solutions 
	% found so far and cut when it reaches N.
	assert(meta:'$found'(ID,0)),
	!, call_cleanup(
		(	call(Query),
			retract(meta:'$found'(ID,J)), succ(J,I),
			(I<N -> assert(meta:'$found'(ID,I)) ; !)
		),
		retractall(meta:'$found'(ID,_))).


% find_nmax(N:natural,X:A,Q:goal,Solns:list(A)).
%
% Like findall/3 but finds upto N solutions.
find_nmax(N,X,Query,XX) :-
	gensym('$find',ID),
	assert(meta:'$found'(ID,list(0,XX,XX))), !,
	!, call_cleanup(
		(	call(Query),
			retract(meta:'$found'(ID,list(J,H,T0))),
			succ(J,I), T0=[X|T1],
			(	I<N 
			-> assert(meta:'$found'(ID,list(I,H,T1))), fail
			;	T1=[], XX=H, !
			)
		;	retract(meta:'$found'(ID,list(J,H,T0))),
			T0=[], XX=H
		),
		retractall(meta:'$found'(ID,_))).
	

% find_n(N:natural,X:A,Q:goal,Solns:list(A)).
%
% Like findall/3 but finds exactly N solutions.
% Fails if N solutions can't be found.
find_n(N,X,Query,XX) :-
	gensym('$find',ID),
	assert(meta:'$found'(ID,list(0,XX,XX))), !,
	!, call_cleanup(
		(	call(Query),
			retract(meta:'$found'(ID,list(J,H,T0))),
			succ(J,I), T0=[X|T1],
			(	I<N
			-> assert(meta:'$found'(ID,list(I,H,T1))), fail
			;	T1=[], XX=H, !
			)
		),
		retractall(meta:'$found'(ID,_))).
	

